# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import paramiko
import socket
import time

import mock

from networking_fujitsu.ml2.fossw import client
from networking_fujitsu.ml2.fossw import mech_fossw
from neutron.tests import base
from oslo_config import cfg

DUMMY_FOSSW_IPS = ['192.168.0.1']
DUMMY_USERNAME = 'username'
MAX_LOGICAL_PORT_NUM = 64


class BaseTestFOSSWClient(base.BaseTestCase):
    """Base class to test Fujitsu Fossw driver."""

    def setUp(self):
        super(BaseTestFOSSWClient, self).setUp()
        cfg.CONF.register_opts(
            mech_fossw.ML2_FUJITSU,
            mech_fossw.ML2_FUJITSU_GROUP
        )
        cfg.CONF.set_override(
            'username', DUMMY_USERNAME, 'fujitsu_fossw')
        cfg.CONF.set_override(
            'fossw_ips', DUMMY_FOSSW_IPS, 'fujitsu_fossw')
        self.cli = client.FOSSWClient(cfg.CONF)
        self.cli.console = mock.Mock()


class TestFOSSWClientConnect(BaseTestFOSSWClient):
    """Test FOSSW client for connect to FOS switch."""

    def setUp(self):
        paramiko.SSHClient = mock.Mock()
        super(TestFOSSWClientConnect, self).setUp()
        self.cli._exec_command = mock.Mock()
        self.ip = cfg.CONF.fujitsu_fossw.fossw_ips[0]

    def test_connect(self):
        self.cli.connect(self.ip)
        self.cli.ssh.connect.assert_called_once_with(
            self.ip,
            password=cfg.CONF.fujitsu_fossw.password,
            port=cfg.CONF.fujitsu_fossw.port,
            timeout=cfg.CONF.fujitsu_fossw.timeout,
            username=cfg.CONF.fujitsu_fossw.username
        )
        self.cli._exec_command.assert_called_once_with(
            client.TERMINAL_LENGTH_0)

    def test_connect_exists_session_and_reuse(self):
        self.cli.ssh = mock.Mock()
        self.cli.ssh._host_keys.lookup.return_value = True
        self.cli._reconnect = mock.Mock()
        self.cli.connect(self.ip)

        self.cli.ssh.connect.assert_not_called()
        self.cli._exec_command.assert_not_called()
        self.cli._reconnect.assert_not_called()

    def test_connect_exists_session_and_not_reuse(self):
        self.cli.ssh = mock.Mock()
        self.cli.ssh._host_keys.lookup.return_value = False
        self.cli.connect(self.ip)

        self.assertEqual(1, self.cli.ssh.connect.call_count)
        self.cli._exec_command.assert_called_once_with(
            client.TERMINAL_LENGTH_0)

    def test_connect_fail(self):
        with mock.patch(__name__ + ".client.paramiko.SSHClient") as p_ssh:
            self.cli._exec_command = mock.Mock()
            errors = [IOError, paramiko.ssh_exception.BadHostKeyException,
                      paramiko.ssh_exception.AuthenticationException,
                      paramiko.ssh_exception.SSHException,
                      socket.error(9999, "fake_error")]
            p_ssh.return_value.connect.side_effect = errors

            self.assertRaises(
                client.FOSSWClientException,
                self.cli.connect,
                cfg.CONF.fujitsu_fossw.fossw_ips[0]
            )
            self.assertIsNone(self.cli.ssh)


class TestFOSSWClientDisconnect(BaseTestFOSSWClient):
    """Test FOSSW client for diconnect from FOS switch."""

    def test_disconnect(self):
        with mock.patch.object(paramiko, 'SSHClient') as p_ssh:
            p_ssh.connect.return_value = None
            self.cli._exec_command = mock.Mock()

            self.cli.connect(cfg.CONF.fujitsu_fossw.fossw_ips[0])
            self.cli.disconnect()
            self.assertIsNone(self.cli.ssh)


class TestFOSSWClientExecCommand(BaseTestFOSSWClient):
    """Test FOSSW client for execute command to FOS switch."""

    def setUp(self):
        super(TestFOSSWClientExecCommand, self).setUp()
        time.sleep = mock.Mock()
        self.cli.console.recv_ready.side_effect = [True, False]
        self.cli.disconnect = mock.Mock()

    def test_normal(self):
        cmd = 'configure'
        self.cli.console.recv.return_value = (
            '\n(ET-7648BRA-FOS) #%s\n(ET-7648BRA-FOS) (Config)#' % cmd)

        result = self.cli._exec_command(cmd)
        self.cli.console.send.assert_called_with(cmd + '\n')
        self.cli.console.recv_ready.assert_called_with()
        self.cli.console.recv.assert_called_once_with(client.RECV_BUF)
        self.assertEqual('(ET-7648BRA-FOS) (Config)#', result)

    def test_socket_timeout_with_send(self):
        cmd = 'configure'
        self.cli.console.send.side_effect = socket.timeout

        self.assertRaises(
            client.FOSSWClientException, self.cli._exec_command, cmd)
        self.cli.disconnect.assert_called_once_with()

    def test_socket_timeout_with_maximum_retry(self):
        self.cli.console.recv_ready = mock.Mock(return_value=False)
        cmd = 'configure'

        self.assertRaises(
            client.FOSSWClientException, self.cli._exec_command, cmd)
        self.cli.disconnect.assert_called_once_with()


class TestFOSSWClientFormatCommand(BaseTestFOSSWClient):
    """Test FOSSW client for format command"""

    def setUp(self):
        super(TestFOSSWClientFormatCommand, self).setUp()

    def test__format_command(self):
        self.assertEqual(
            "vlan 2", self.cli._format_command("vlan {vlanid}", vlanid=2))

    def test__format_command_fails(self):
        raw_cmd = "vlan {vlanid}"

        self.assertRaises(
            client.FOSSWClientException, self.cli._format_command, raw_cmd)
        self.assertRaises(
            client.FOSSWClientException, self.cli._format_command, raw_cmd,
            vlan=2)
        self.assertRaises(
            client.FOSSWClientException, self.cli._format_command, raw_cmd,
            vlanid=None)


class TestFOSSWClientChangeMode(BaseTestFOSSWClient):
    """Test FOSSW client for format command"""

    def setUp(self):
        super(TestFOSSWClientChangeMode, self).setUp()
        self.cli._exec_command = mock.Mock(return_value=') #')

    def test_non_admin_mode_after_end(self):
        self.cli._exec_command.return_value = ') >'

        self.cli.change_mode('configure')
        self.assertEqual(3, self.cli._exec_command.call_count)
        actual = [arg[0][0] for arg in self.cli._exec_command.call_args_list]
        self.assertEqual(['end', 'enable', 'configure'], actual)

    def test_mode_global(self):
        self.cli.change_mode('configure')

        self.assertEqual(2, self.cli._exec_command.call_count)
        actual = [arg[0][0] for arg in self.cli._exec_command.call_args_list]
        self.assertEqual(['end', 'configure'], actual)

    def test_mode_vlan(self):
        self.cli.change_mode('vlan database')

        self.assertEqual(2, self.cli._exec_command.call_count)
        actual = [arg[0][0] for arg in self.cli._exec_command.call_args_list]
        self.assertEqual(['end', 'vlan database'], actual)

    def test_mode_interface_with_ifname(self):
        self.cli.change_mode('interface', ifname='0/30')

        self.assertEqual(3, self.cli._exec_command.call_count)
        actual = [arg[0][0] for arg in self.cli._exec_command.call_args_list]
        self.assertEqual(['end', 'configure', 'interface 0/30'], actual)

    def test_illegal_mode_interface_only(self):
        self.assertRaises(
            client.FOSSWClientException,
            self.cli.change_mode, 'interface')

    def test_illegal_undefined_mode(self):
        self.cli.change_mode('undefined')
        self.cli._exec_command.assert_called_once_with(client.END)


class TestFOSSWClientCreateVlan(BaseTestFOSSWClient):
    """Test Fossw client for create VLAN"""
    def setUp(self):
        super(TestFOSSWClientCreateVlan, self).setUp()
        self.cli.change_mode = mock.Mock(return_value=None)
        self.cli._exec_command = mock.Mock()

    def test_create_vlan(self):
        result = "(ET-7648BRA-FOS) (Vlan)#"
        self.cli._exec_command.return_value = "(ET-7648BRA-FOS) (Vlan)#"
        self.assertEqual(result, self.cli.create_vlan(10))


class TestFOSSWClientDeleteVlan(BaseTestFOSSWClient):
    """Test Fossw client for delete VLAN"""
    def setUp(self):
        super(TestFOSSWClientDeleteVlan, self).setUp()
        self.cli.change_mode = mock.Mock(return_value=None)
        self.cli._exec_command = mock.Mock()

    def test_delete_vlan(self):
        ret = "(ET-7648BRA-FOS) (Vlan)#"
        self.cli._exec_command.return_value = ret
        self.assertIsNone(self.cli.delete_vlan(10))

    def test_delete_vlan_already_deleted(self):
        ret = ("Failed to delete one or more VLAN's. Please refer system log "
               "for more information.")
        self.cli._exec_command.return_value = ret
        self.assertIsNone(self.cli.delete_vlan(10))


class TestFOSSWClientSetVlan(BaseTestFOSSWClient):
    """Test Fossw client for set VLAN"""
    def setUp(self):
        super(TestFOSSWClientSetVlan, self).setUp()
        self.cli.change_mode = mock.Mock(return_value=None)
        self.cli._exec_command = mock.Mock()

    def test_set_vlan(self):
        ret = ["(ET-7648BRA-FOS) (Interface 0/1)#",
               "(ET-7648BRA-FOS) (Interface 0/1)#"]
        self.cli._exec_command.side_effect = ret
        self.assertIsNone(self.cli.set_vlan(10, 1))

    def test_set_vlan_fail(self):
        ret = ["(ET-7648BRA-FOS) (Interface 0/1)#",
               "VLAN ID not found.(ET-7648BRA-FOS) (Interface 0/1)#"]
        self.cli._exec_command.side_effect = ret
        self.assertRaises(
            client.FOSSWClientException, self.cli.set_vlan, 10, 1)


class TestFOSSWClientClearVlan(BaseTestFOSSWClient):
    """Test Fossw client for clear VLAN"""
    def setUp(self):
        super(TestFOSSWClientClearVlan, self).setUp()
        self.cli.change_mode = mock.Mock(return_value=None)
        self.cli._exec_command = mock.Mock()

    def test_clear_vlan(self):
        ret = "(ET-7648BRA-FOS) (Interface 0/1)#"
        self.cli._exec_command.side_effect = ret
        self.assertIsNone(self.cli.clear_vlan(1))


class TestFOSSWClientJoinToLag(BaseTestFOSSWClient):
    """Test FOSSW Client for join to lag"""
    def setUp(self):
        super(TestFOSSWClientJoinToLag, self).setUp()
        self.cli.change_mode = mock.Mock(return_value=None)
        self.cli._exec_command = mock.Mock(return_value=None)

    def test_join_to_lag(self):
        ret = ["(ET-7648BRA-FOS) (Interface 3/1)#",
               "(ET-7648BRA-FOS) (Interface 0/1)#"]
        self.cli._exec_command.side_effect = ret
        self.assertIsNone(self.cli.join_to_lag("0/1", "3/1"))


class TestFOSSWClientJoinToVPC(TestFOSSWClientJoinToLag):
    """Test FOSSW Client for join to vpc"""

    def test_join_to_vpc(self):
        logicalport = '1'
        vpcid = '10'
        self.cli.join_to_vpc(logicalport, vpcid)
        self.cli.change_mode.assert_called_once_with(
            client.MODE_INTERFACE, ifname=logicalport)
        self.cli._exec_command.assert_called_once_with(
            "vpc {vpcid}".format(vpcid=vpcid))


class TestFOSSWClientGetVPCId(BaseTestFOSSWClient):
    """TEst FOSSW Client for get vpc id"""
    def setUp(self):
        super(TestFOSSWClientGetVPCId, self).setUp()
        self.cli._exec_command = mock.Mock()

    def test_get_vpc_specified_id(self):
        ret = [
            "Port channel................................... none",
            "Port channel................................... none",
            "Port channel................................... none",
            "Port channel................................... 3/1",
            "Port channel................................... 3/2",
            "Port channel................................... none"]
        self.cli._exec_command.side_effect = ret
        self.assertEqual("5", self.cli.get_vpcid(logicalport="3/2"))

    def test_get_vpc_free_id(self):
        ret = [
            "Port channel................................... none",
            "Port channel................................... none",
            "Port channel................................... none"]
        self.cli._exec_command.side_effect = ret
        self.assertEqual("1", self.cli.get_vpcid())

        ret[0] = "Port channel................................... 3/1"
        self.cli._exec_command.side_effect = ret
        self.assertEqual("2", self.cli.get_vpcid())

    def test_get_vpc_id_no_free_vpc(self):
        ret = []
        for i in iter(range(1, MAX_LOGICAL_PORT_NUM + 1)):
            ret.append(
                ("Port channel................................... "
                 "3/{no}").format(no=i))
        self.cli._exec_command.side_effect = ret
        self.assertIsNone(self.cli.get_vpcid())

    def test_get_vpc_specidied_id_not_found(self):
        ret = "Port channel................................... none"
        self.cli._exec_command.return_value = ret
        self.assertIsNone(self.cli.get_vpcid("3/1"))


class TestFOSSWClientGetPeerlinkPartner(BaseTestFOSSWClient):
    """Test FOSSW Client for get peerlink partner ip"""
    def setUp(self):
        super(TestFOSSWClientGetPeerlinkPartner, self).setUp()
        self.cli._exec_command = mock.Mock()

    def test_get_peerlink_partner(self):
        self.cli._exec_command.return_value = (
            "Peer IP address................................ 192.168.1.1")
        self.assertEqual("192.168.1.1", self.cli.get_peerlink_partner())


class TestFOSSWClientGetLAGPort(BaseTestFOSSWClient):
    """Test FOSSW Client for get port-wqmemberport of port-channel."""

    def setUp(self):
        super(TestFOSSWClientGetLAGPort, self).setUp()
        self.cli._exec_command = mock.Mock()

    def test_lag_lagport_found_with_portname(self):
        ret = ("addport 3/1\n"
               "exit\n"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertEqual("3/1", self.cli.get_lag_port('0/1'))

    def test_lag_lagport_found_without_portname(self):
        ret = ("3/1   ch1   1   Down  Disabled Static 0/1,0/2\n"
               "3/2   ch2   1   Down  Disabled Static 0/3,0/4\n"
               "3/3   ch3   1   Down  Disabled Static\n"
               "3/4   ch4   1   Down  Disabled Static 0/5,0/6\n"
               "3/5   ch5   1   Down  Disabled Static  \n"
               "3/6   ch6   1   Down  Disabled Static\n"
               "3/7   ch7   1   Down  Disabled Static"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        # Returns earlier number of lagport
        self.assertEqual("3/3", self.cli.get_lag_port())

    def test_lag_all_lagport_are_dynamic_with_portname(self):
        ret = ("addport 3/2\n"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertEqual('3/2', self.cli.get_lag_port('0/3'))

    def test_lag_all_lagport_are_dynamic_without_portname(self):
        ret = ("3/1   ch1   1   Down  Disabled Dynamic 0/1,0/2\n"
               "3/2   ch2   1   Down  Disabled Dynamic 0/3,0/4\n"
               "3/3   ch3   1   Down  Disabled Dynamic\n"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertIsNone(self.cli.get_lag_port())
        self.cli._exec_command.assert_called_once_with(
            'show port-channel brief | exclude Dynamic begin 3/')

    def test_mlag_found_with_portname(self):
        ret = ("addport 3/2\n"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertEqual('3/2', self.cli.get_lag_port('0/3'))
        self.cli._exec_command.assert_called_once_with(
            'show running-config interface 0/3')

    def test_mlag_found_without_portname(self):
        ret = ("3/1   ch1   1   Down  Disabled Dynamic 0/1\n"
               "3/2   ch2   1   Down  Disabled Dynamic 0/3\n"
               "3/3   ch3   1   Down  Disabled Static\n"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertEqual('3/3', self.cli.get_lag_port())
        self.cli._exec_command.assert_called_once_with(
            'show port-channel brief | exclude Dynamic begin 3/')

    def test_mlag_not_found_with_portname(self):
        ret = ("\n(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertIsNone(self.cli.get_lag_port())

    def test_illegal_duplicate_target_found_with_portname(self):
        ret = ("addport 3/1 \n"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        # Returns earlier number of lagport
        self.assertEqual("3/1", self.cli.get_lag_port('0/1'))
        self.cli._exec_command.assert_called_once_with(
            'show running-config interface 0/1')

    def test_illegal_duplicate_target_found_without_portname(self):
        ret = ("3/2   ch2   1   Down  Disabled Static  0/1,0/2\n"
               "3/3   ch3   1   Down  Disabled Static   \n"
               "3/4   ch4   1   Down  Disabled Static\n"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        # Returns earlier lagport
        self.assertEqual("3/3", self.cli.get_lag_port())

    def test_illegal_static_lag_port_exists(self):
        ret = ("3/1   ch1   1   Down  Disabled Static  0/1,0/2,\n"
               "3/2   ch2   1   Down  Disabled Static 0/5,0/6\n"
               "3/3   ch3   1   Down  Disabled Dynamic\n"
               "3/4   ch4   1   Down  Disabled Static \n"
               "3/5   ch5   1   Down  Disabled Static\n"
               "(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        # Returns earlier lagport
        self.assertEqual("3/4", self.cli.get_lag_port())


class TestFOSSWClientGetSwitchMAC(BaseTestFOSSWClient):
    """Test FOSSW Client for get FOS switch's MAC address"""
    def setUp(self):
        super(TestFOSSWClientGetSwitchMAC, self).setUp()
        self.cli._exec_command = mock.Mock()

    def test_returns_normal(self):
        ret = ("Base MAC Address............................... "
               "00:30:AB:F4:CA:DA(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertEqual('00:30:ab:f4:ca:da', self.cli.get_switch_mac())

    def test_returns_mac_and_any_return_codes(self):
        ret = ("\n\n00:30:AB:F4:CA:DA\n\n")
        self.cli._exec_command.return_value = ret
        self.assertEqual('00:30:ab:f4:ca:da', self.cli.get_switch_mac())

    def test_returns_mac_all_capital(self):
        ret = ("\n\nAA:AA:AA:AA:AA:AA\n\n")
        self.cli._exec_command.return_value = ret
        self.assertEqual('aa:aa:aa:aa:aa:aa', self.cli.get_switch_mac())

    def test_returns_mac_all_number(self):
        ret = ("Base MAC Address............................... "
               "00:30:22:44:55:66(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertEqual('00:30:22:44:55:66', self.cli.get_switch_mac())

    def test_returns_none(self):
        self.cli._exec_command.return_value = None
        self.assertIsNone(self.cli.get_switch_mac())

    def test_returns_illegal_output(self):
        ret = ("System is Busy. Please execute again................")
        self.cli._exec_command.return_value = ret
        self.assertIsNone(self.cli.get_switch_mac())

    def test_returns_illegal_mac(self):
        ret = ("Base MAC Address............................... "
               "GG:GG:GG:GG:GG:GG(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertIsNone(self.cli.get_switch_mac())

    def test_returns_illegal_mac_long(self):
        ret = ("Base MAC Address............................... "
               "00:30:AB:F4:CA:DA:8B(ET-7648BRA-FOS) #")
        self.cli._exec_command.return_value = ret
        self.assertEqual('00:30:ab:f4:ca:da', self.cli.get_switch_mac())


class TestFOSSWClientLeaveFromLAG(BaseTestFOSSWClient):
    """Test FOSSW Client for leave physical port from logical port"""
    def setUp(self):
        super(TestFOSSWClientLeaveFromLAG, self).setUp()
        self.cli.change_mode = mock.Mock(return_value=None)
        self.cli._exec_command = mock.Mock()

    def test_leave_from_lag(self):
        ret = "(ET-7648BRA-FOS) (Interface 0/10)#"
        self.cli._exec_command.return_value = ret
        self.assertIsNone(self.cli.leave_from_lag("0/10", "3/1"))


class TestFOSSWClientLeaveFromVPC(BaseTestFOSSWClient):
    """Test FOSSW Client for leave logical port from VPC(mLAG)"""
    def setUp(self):
        super(TestFOSSWClientLeaveFromVPC, self).setUp()
        self.cli.change_mode = mock.Mock(return_value=None)
        self.cli._exec_command = mock.Mock()

    def test_leave_from_vpc(self):
        ret = ["(ET-7648BRA-FOS) (Interface 3/1)#",
               "          Failure Information"
               "---------------------------------------"
               "Interfaces failed to be configured : 1"
               "---------------------------------------"
               "Interface             Error"
               "---------------------------------------"
               "3/1           Failed to remove."
               "(ET-7648BRA-FOS) (Interface 3/1)#"]
        self.cli._exec_command.side_effect = ret
        self.assertIsNone(self.cli.leave_from_vpc("3/1", "1"))
        self.assertIsNone(self.cli.leave_from_vpc("3/1", "2"))


class TestFOSSWClientSaveRunningConfig(BaseTestFOSSWClient):
    """Test FOSSW client for save running config."""
    def setUp(self):
        super(TestFOSSWClientSaveRunningConfig, self).setUp()
        self.cli._exec_command = mock.Mock(return_value="(ET-7648BRA-FOS) #")

    def test_save_running_config(self):
        self.assertIsNone(self.cli.save_running_config())



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import paramiko
import re
import socket
import time

from oslo_log import log as logging

from networking_fujitsu._i18n import _
from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LW

LOG = logging.getLogger(__name__)

MODE_GLOBAL = 'configure'
MODE_VLAN = 'vlan database'
MODE_INTERFACE = 'interface'
ENABLE = 'enable'
END = 'end'
READ_TIMEOUT = 5.0
MAX_LOOP = 50
MAX_VPC_ID = 63
SHOW_PC_BRIEF = 'show port-channel brief'
TERMINAL_LENGTH_0 = 'terminal length 0'
RECV_BUF = 8192


class FOSSWClient(object):

    def __init__(self, conf):
        self._conf = conf
        self.ssh = None
        self.console = None

    def connect(self, ip):
        """Establish ssh connection for FOS switch.

        If exists session for specified target, reuse it otherwise reconnect.
        In order to get command result correctly, execute 'terminal length 0'
        at the beginning of the session.
        :param ip: a ip address of FOS switch
        :type ip: string
        :returns: None
        :rtype: None

        """
        if not self.lookup(ip):
            self._reconnect(ip)
            self._exec_command(TERMINAL_LENGTH_0)

    def _reconnect(self, ip):
        """Reconnect a new SSH session

        :param ip: a ip address of FOS switch
        :type ip: string
        :returns: None
        :rtypes: None

        """
        self.disconnect()
        method = "connect"
        retry_count = 0
        while retry_count < 5:
            try:
                self.ssh = paramiko.SSHClient()
                self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self.ssh.connect(
                    ip,
                    port=self._conf.fujitsu_fossw.port,
                    username=self._conf.fujitsu_fossw.username,
                    password=self._conf.fujitsu_fossw.password,
                    timeout=self._conf.fujitsu_fossw.timeout
                )
                self.console = self.ssh.invoke_shell()
                self.console.settimeout(READ_TIMEOUT)
                return
            except IOError as e:
                LOG.warning(_LW('Could not initialize SSH client. %s'), e)
            except (paramiko.ssh_exception.BadHostKeyException,
                    paramiko.ssh_exception.AuthenticationException,
                    paramiko.ssh_exception.SSHException) as e:
                LOG.warning(_LW('Could not connect to FOS switch. An error'
                                'occurred while connecting. %s'), e)
            except socket.error as e:
                e_no, e_str = e
                LOG.warning(_LW('A socket error occurred while connecting.\n'
                                '[Errno %(e_no)s] %(e_str)s'),
                            {'e_no': e_no, 'e_str': e_str})
            except Exception as e:
                LOG.warning(_LW('Unexpected error occurred while connecting. '
                                '%s'), e)
            retry_count += 1
            self.disconnect()
            LOG.warning(_LW('Connect attempt %s failed.'), retry_count)
        if retry_count >= 5:
            self.disconnect()
            LOG.exception(_LE('Max retries exceeded. Failed to connect to '
                              'FOS switch.'))
            raise FOSSWClientException(method)

    def disconnect(self):
        """Close ssh connection with FOS switch.

        :returns: None
        :rtypes: None

        """
        if self.ssh:
            self.ssh.close()
            self.ssh = None
            self.console = None
        else:
            LOG.debug(_("SSH connection already closed."))

    def _exec_command(self, command):
        try:
            raw_res = ""
            self.console.send(command + "\n")
            LOG.debug(_("FOSSW client sent: %s"), command)
            i = 0
            while i < MAX_LOOP:
                time.sleep(0.1)
                if self.console.recv_ready():
                    raw_res += self.console.recv(RECV_BUF)
                elif raw_res:
                    break
                i += 1
            if (i == MAX_LOOP):
                LOG.error(_LE("No reply from FOS switch."))
                raise socket.timeout
            index = raw_res.find(command)
            received = raw_res[(index + len(command) + 1):]
        except socket.timeout:
            self.disconnect()
            LOG.exception(_LE('Socket timeout occured while executing '
                              'commands to FOS Switch.'))
            raise FOSSWClientException('_exec_command')
        else:
            LOG.debug(_("FOSSW client received: %s"), received)
            # NOTE(yushiro) Validate received message here
            return received.replace('\r\n', '\n')

    def _format_command(self, command, **kwargs):
        method = "_format_command"
        if not all(kwargs.values()):
            raise FOSSWClientException(method)
        try:
            formatted = command.format(**kwargs)
        except (KeyError, TypeError):
            raise FOSSWClientException(method)
        return formatted

    def lookup(self, target):
        """Check exist session for specified IP.

        :param target: IP address of a target host
        :type target: string

        :returns: exist(True) or not exist(False)
        :rtype: bool

        """
        if self.ssh:
            if self.ssh._host_keys:
                if self.ssh._host_keys.lookup(target):
                    LOG.debug('Reuse existing session for %s.', target)
                    return True
        return False

    def create_vlan(self, segmentation_id):
        """Define VLAN with specified VLAN ID to FOS switch.

        :param segmentation_id: id of VLAN to be created.
        :type segmentation_id: string

        :returns: received message from FOS switch
        :rtype: string

        """
        self.change_mode(MODE_VLAN)
        cmd = self._format_command("vlan {vlanid}", vlanid=segmentation_id)
        return self._exec_command(cmd)

    def delete_vlan(self, segmentation_id):
        """Undefine VLAN with specified VLAN ID from FOS switch.

        :param segmentation_id: id of VLAN to be deleted.
        :type segmentation_id: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_VLAN)
        cmd = self._format_command("no vlan {vlanid}", vlanid=segmentation_id)
        if "Failed to delete" in self._exec_command(cmd):
            LOG.warning(_LW("VLAN(%s) has already deleted."), segmentation_id)

    def set_vlan(self, segmentation_id, port_id):
        """Associate VLAN to specified physical port on FOS switch.

        :param segmentation_id: id of VLAN to be associated
        :type segmentation_id: string
        :param port_id: the number of physical port on FOS switch
        :type port_id: string

        :returns: None
        :rtype: None

        """
        method = "set_vlan"
        self.change_mode(MODE_INTERFACE, port_id)
        self._exec_command("switchport mode access")
        cmd = self._format_command("switchport access vlan {vlanid}",
                                   vlanid=segmentation_id)
        if "VLAN ID not found." in self._exec_command(cmd):
            LOG.exception(_LE("VLAN(%s) does not exist on FOS switch. "
                              "Please check vlan setting."), segmentation_id)
            raise FOSSWClientException(method)

    def clear_vlan(self, port_id):
        """Disassociate VLAN from specified physical port on FOS switch.

        :param port_id: the number of physical port on FOS switch
        :type port_id: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, port_id)
        self._exec_command("no switchport access vlan")
        self._exec_command("no switchport mode")

    def join_to_lag(self, port, logicalport):
        """Join specified port to LAG configuration.

        :param port: the number of physical port on FOS switch
        :type port: string
        :param logicalport: the number of logical port on FOS switch
        :type logicalport: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, ifname=logicalport)
        self._exec_command("no port-channel static")

        self.change_mode(MODE_INTERFACE, ifname=port)
        self._exec_command("addport {lo_port}".format(lo_port=logicalport))

    def get_vpcid(self, logicalport="none"):
        """Get VPC ID which is associated with specified logical port.

        If this method is called without logicalport, this method returns a
        VPC ID which is not associated with any logicalport.

        :param logicalport: the number of logicalport
            (optional, default: "none")
        :type logicalport: string

        :returns: the ID of VPC or None.
        :rtype: string
        """
        # TODO(yushiro): Replace regexp
        for i in iter(range(1, MAX_VPC_ID + 1)):
            cmd = 'show vpc {vid} | include "Port channel"'.format(vid=str(i))
            tmp_text = self._exec_command(cmd)
            switch_logicalport = tmp_text[tmp_text.find('. ') + 2:]
            if switch_logicalport == logicalport:
                return str(i)
        if logicalport is "none":
            # NOTE(miyagishi_t): All vpc is already associated to any logical
            # port. Therefore, the FOS switch cannot to define a new vpc.
            LOG.error(_LE("There is no free vpc. All vpc is already "
                          "configured."))
        else:
            # NOTE(miyagishi_t): There is no VPC id which is related to
            # specified logicalport. It maybe the vpc already cleard or
            # originaly not defined on the FOS switch.
            LOG.warning(_LW("A vpc which related to logicalport(%s) on FOS "
                            "switch not found."), logicalport)

    def join_to_vpc(self, logicalport, vpcid):
        """Join a specified logical port to member of VPC.

        :param logicalport: the number of logical port on FOS switch
        :type logicalport: string
        :param vpcid: id of VPC
        :type vpcid: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, ifname=logicalport)
        self._exec_command("vpc {vpcid}".format(vpcid=vpcid))

    def get_peerlink_partner(self):
        """Get peerlink partner switch's IP address.

        :returns: IP address of peerlink partner
        :rtype: string

        """
        # TODO(yushiro): Replace regexp
        ret = self._exec_command('show vpc peer-keepalive | include "Peer '
                                 'IP address"')
        return ret[ret.find('. ') + 2:]

    def get_lag_port(self, portname=None):
        """Get logicalport from FOS switch.

        If portname is specified, returns logicalport which is associated
        with portname. If portname is not specified, returns available
        logicalport in FOS switch. 'available' means as follows:
            1. Status is not 'Dynamic'
            2. Portname Begins '3/'
            3. No member port exists
        :param portname: the number of physical port on FOS switch
        :type portname: string
        :returns: the number of logical port which associated with specified
                  physical port
        :rtype: string

        """
        # Find related lag_port with specified portname
        if portname:
            cmd = "show running-config interface %s" % portname
            query = r'addport\s+(?:(\d+/(?:(\d+))))'
        # Find free lag_port from all port-channel
        else:
            cmd = SHOW_PC_BRIEF + " | exclude Dynamic begin 3/"
            query = r'(?:(3/(?:(\d+))).+Static(?!\s+\d+\/\d+,\d+\/\d+))'
        ret = re.search(query, self._exec_command(cmd))
        return ret.group(1) if ret else None

    def get_switch_mac(self):
        """Get MAC address of FOS switch.

        :returns: MAC address of FOS switch
        :rtype: string

        """
        cmd = 'show hardware eeprom | include "Base MAC Address"'
        res = self._exec_command(cmd)
        if res:
            mac = re.search(r'([0-9A-F]{2}(:[0-9A-F]{2}){5})', res)
            return mac.group(0).lower() if mac else None

    def leave_from_lag(self, port, logicalport):
        """Leave a specified port from LAG configuration.

        :param port: the number of physical port on FOS switch
        :type port: string
        :param logicalport: the number of logical port on FOS switch
        :type logicalport: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, ifname=port)
        res = self._exec_command(
            "deleteport {lo_port}".format(lo_port=logicalport))
        self.change_mode(MODE_INTERFACE, ifname=logicalport)
        self._exec_command("port-channel static")
        if "is not a member of port-channel" in res:
            LOG.warning(_LW("Specified port(%(port)s) has already removed "
                            "from logical port(%(log_po)s)"),
                        {"port": port, "log_po": logicalport})

    def leave_from_vpc(self, logicalport, vpcid):
        """Leave a specified logical port from member of VPC.

        :param logicalport: the number of logical port on FOS switch
        :type logicalport: string
        :param vpcid: id of VPC
        :type vpcid: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, ifname=logicalport)
        res = self._exec_command("no vpc {vpcid}".format(vpcid=vpcid))
        if "Failed to remove" in res:
            LOG.warning(_LW("Specified logical port(%(log_po)s) has already "
                            "removed from VPC(%(vpc)s)."),
                        {"log_po": logicalport, "vpc": vpcid})

    def save_running_config(self):
        """Save running config of FOS switch.

        :returns: None
        :rtypes: None

        """
        self._exec_command("copy system:running-config nvram:startup-config")
        self._exec_command("y")
        self._exec_command("")

    def change_mode(self, mode, ifname=None):
        """Change CLI mode of FOS switch.

        :param mode: the name of the mode to be moved.
        :type mode: string
        :param ifname: the number of physical port on FOS switch
            (optional, default: None)
        :type ifname: string

        :returns: None
        :rtype: None

        """
        # Move to Privileged EXEC mode.
        prompt = self._exec_command(END)
        if ") >" in prompt:
            self._exec_command(ENABLE)
        if mode in [MODE_GLOBAL, MODE_VLAN]:
            self._exec_command(mode)
        elif mode == MODE_INTERFACE:
            self._exec_command(MODE_GLOBAL)
            cmd = self._format_command(
                MODE_INTERFACE + " {ifname}", ifname=ifname)
            self._exec_command(cmd)


class FOSSWClientException(Exception):

    def __init__(self, method):
        self.message = _("FOSSW client caught an error at %s") % method

    def __str__(self):
        return(self.message)



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from oslo_config import cfg
from oslo_log import log as logging

from networking_fujitsu._i18n import _LI
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.common.ovsdb import ovsdb_writer
from networking_fujitsu.ml2.common import tunnel_caller
from networking_fujitsu.ml2.common import type_vxlan
from networking_fujitsu.ml2.fossw import client

from neutron_lib import context

from neutron.common import utils

LOG = logging.getLogger(__name__)
_LOCK_NAME = 'fujitsu_fossw'


class FOSSWVxlanDriver(object):
    """Fossw CLI interface driver for Neutron network.

    Handles life-cycle management of Neutron network (OVSDB on FOSSW)
    """

    def __init__(self, conf=None):
        if conf:
            self._conf = conf
        else:
            self._conf = cfg.CONF
        self.ovsdb_port = self._conf.fujitsu_fossw.ovsdb_port
        self.udp_dest_port = self._conf.fujitsu_fossw.udp_dest_port
        self.ovsdb_vlanid_range_min = (
            self._conf.fujitsu_fossw.ovsdb_vlanid_range_min)
        self.fossw_ips = self._conf.fujitsu_fossw.fossw_ips

        self.client = client.FOSSWClient(self._conf)
        self.type_vxlan = type_vxlan.TypeVxlan()
        self.vxlan_endpoint_ips = []
        self.tunnel_caller = tunnel_caller.TunnelCaller()
        self.initialize()

    def initialize(self):
        """Initialize update

        Both update Neutron DB with switch VTEPs, and update FOS switches
        with host & other switches' VTEPs.
        """

        if "vxlan" in self._conf.ml2.type_drivers:
            self._update_neutron_db()
            self.vxlan_endpoint_ips = [vxlan_endpoint['ip_address']
                                       for vxlan_endpoint in
                                       self.type_vxlan.get_endpoints()]

    def save_all_fossw(self):
        """Save running-config to startup-config in all FOS switches."""
        for fossw_ip in self.fossw_ips:
            self.client.connect(fossw_ip)
            self.client.save_running_config()
            self.client.disconnect()

    def _update_neutron_db(self):
        """Update Neutron DB ml2_vxlan_endpoints table

        Update with VTEPs of all FOS switches.
        :returns: None
        """

        for fossw_ip in self.fossw_ips:
            ovsdb_cli = ovsdb_writer.OVSDBWriter(fossw_ip, self.ovsdb_port)
            sw_ep_ip, sw_ep_host = ovsdb_cli.get_sw_ep_info()
            if not sw_ep_host:
                sw_ep_host = "FOSSW"
            if sw_ep_ip:
                neutron_db = self.type_vxlan.get_endpoint_by_ip(
                    sw_ep_ip)
                if not neutron_db:
                    LOG.info(_LI("Add FOS switch endpoint information (ip = "
                                 "%(ip)s, hostname = %(name)s, udp_dest_port "
                                 "= %(port)s) to ml2_vxlan_endpoints table in "
                                 "Neutron DB."), {'ip': sw_ep_ip,
                                                  'name': sw_ep_host,
                                                  'port': self.udp_dest_port})
                    self.type_vxlan.add_endpoint(sw_ep_ip, sw_ep_host,
                                                 self.udp_dest_port)
                elif (neutron_db['host'] != sw_ep_host or
                      neutron_db['udp_port'] != self.udp_dest_port):
                    LOG.warning(_LW("Existed endpoint information with same "
                                    "IP address (%(ip)s) has been updated "
                                    "with hostname = %(host)s, udp_dest_port "
                                    "= %(port)s."),
                                {'ip': sw_ep_ip, 'host': sw_ep_host,
                                 'port': self.udp_dest_port})
                    self.type_vxlan.delete_endpoint(sw_ep_ip)
                    self.type_vxlan.add_endpoint(sw_ep_ip, sw_ep_host,
                                                 self.udp_dest_port)
                else:
                    LOG.info(_LI("FOS switch endpoint information (ip = "
                                 "%(ip)s, hostname = %(name)s, udp_dest_port "
                                 "= %(port)s) has alreadly been added to "
                                 "ml2_vxlan_endpoints table in Neutron DB. "
                                 "Do nothing."),
                             {'ip': sw_ep_ip, 'name': sw_ep_host,
                              'port': self.udp_dest_port})
            else:
                LOG.warning(_LW("Unable to get endpoint information from "
                                "switch (IP = %s). Skip."), fossw_ip)

    @utils.synchronized(_LOCK_NAME, external=True)
    def create_logical_switch(self, net_uuid, vni, save=True):
        """Create a row in Logical_Switch table in FOS switch OVSDB.

        We consider net_uuid is always unique, and both vni and net_uuid
        must be immutable. So there is not a update case.
        :param net_uuid: The uuid of Neutron network.
        :type net_uuid: string
        :param vni: The segment ID of Neutron network.
        :type vni: integer

        :returns: None
        """

        for fossw_ip in self.fossw_ips:
            ovsdb_cli = ovsdb_writer.OVSDBWriter(fossw_ip, self.ovsdb_port)
            ovsdb_cli.insert_logical_switch(vni, net_uuid.replace("-", ""))
        if save:
            self.save_all_fossw()

    @utils.synchronized(_LOCK_NAME, external=True)
    def delete_logical_switch(self, net_uuid, save=True):
        """Delete row in Logical_Switch table in FOS switch OVSDB.

        We only consider deleting the existing Logical_Switch ROW.
        Also note that even it does not exit, the jsonrpc returns no error.
        :param net_uuid: The uuid of Neutron network.
        :type: string

        :returns: None
        """

        ls_name = net_uuid.replace("-", "")
        for fossw_ip in self.fossw_ips:
            ovsdb_cli = ovsdb_writer.OVSDBWriter(fossw_ip, self.ovsdb_port)
            ls_uuid = ovsdb_cli.get_logical_switch_uuid(ls_name)
            if ls_uuid:
                ovsdb_cli.delete_logical_switch(ls_uuid)
        if save:
            self.save_all_fossw()

    def update_physical_port(self, vni, lli, port, ip_mac_pairs, req_id=None,
                             save=True):
        """Update Physical_Port table in FOS switch OVSDB.

        There are 3 cases about port operation
        Case 1:
        Port in FOS switch need to get bind. In this case, Ucast_Macs_Local
        table and Physical_locator table also need to insert a new row or
        update at the same time.
        Case 2:
        Another FOS switch has updated its port. In this case, other FOS
        switches need to sync the information to Ucast_Macs_Remote table and
        Physical_locator table.
        Case 3:
        Nova Compute Node has created a new VM or something others occurred
        around Neutron network. In this case, all FOS switches should do the
        same thing as in Case 2.

        :param vni: The segment ID of Neutron network which the port belongs
                     to.
        :type vni: integer
        :param lli: The local_link_information of the port.
                    If it is a VM port, then [{}] should be given.
        :type lli: list
        :param port: Dictionary of a port
        :type port: dictionary
        :param ip_mac_pairs: List of MAC(key) - IP(value) pairs of all FOS
                             switches.
        :type ip_mac_pairs: dictionary

        :returns: None
        """

        ls_name = port["network_id"].replace("-", "")
        mac = port["mac_address"]
        port_ips = [fixed_ip['ip_address'] for fixed_ip in port["fixed_ips"]]
        host_id = port['binding:host_id']
        target_ip = ip_mac_pairs[lli[0]['switch_id']] if lli else None
        target = lli[0]['switch_info'] if lli else host_id
        tunnel_ip = self.type_vxlan.db_get_endpoint_ip_by_host(target)

        if lli:
            sw_port = lli[0]['port_id']
            # Update Physical_Port table first.
            ovsdb_cli = ovsdb_writer.OVSDBWriter(target_ip, self.ovsdb_port)
            bind_ls_uuid = ovsdb_cli.get_logical_switch_uuid(ls_name)
            binding_vid = ovsdb_cli.get_binding_vid(bind_ls_uuid)
            bind_vid = binding_vid if binding_vid else (
                int(sw_port[2:]) + self.ovsdb_vlanid_range_min - 1)

            ovsdb_cli.update_physical_port(sw_port, bind_vid, bind_ls_uuid)
            # After Physical_Port table has been updated, update
            # Ucast_Macs_Local table.
            # If any garbage exist, remove them first.
            if ovsdb_cli.get_ucast_macs_local(mac):
                ovsdb_cli.delete_ucast_macs_local(mac)
            # Then add what we want to the Ucast_Macs_Local table.
            locator_ip_local, sw_ep_host = ovsdb_cli.get_sw_ep_info()
            locator_uuid_local = ovsdb_cli.get_physical_locator_uuid(
                locator_ip_local)
            # We do not care about port IP address. We only forward using port
            # MAC address.
            if locator_uuid_local:
                ovsdb_cli.insert_ucast_macs_local(
                    bind_ls_uuid, locator_uuid_local, mac)
            else:
                ovsdb_cli.insert_ucast_macs_local_and_locator(
                    bind_ls_uuid, locator_ip_local, mac)
            # Create vxlan port for FOS Switch when this method was called with
            # req_id.
            if req_id:
                ctxt = context.Context(request_id=req_id, is_admin=True)
                self.tunnel_caller.trigger_tunnel_sync(ctxt, target_ip)
        # At last Ucast_Macs_Remote table of other switches.
        self._update_ucast_macs_remote(
            target_ip, ls_name, mac, tunnel_ip, port_ips)
        if save:
            self.save_all_fossw()

    def _update_ucast_macs_remote(self, target_ip, logical_switch_name, mac,
                                  tunnel_ip, port_ips):
        """Update Ucast_Macs_Remote table in all FOS switches OVSDB."""
        for fossw_ip in self.fossw_ips:
            if target_ip == fossw_ip:
                continue
            ovsdb_cli = ovsdb_writer.OVSDBWriter(fossw_ip, self.ovsdb_port)
            ls_uuid = ovsdb_cli.get_logical_switch_uuid(logical_switch_name)
            if ovsdb_cli.get_ucast_macs_remote(mac):
                ovsdb_cli.delete_ucast_macs_remote(mac)
            locator_uuid = ovsdb_cli.get_physical_locator_uuid(tunnel_ip)
            if locator_uuid:
                ovsdb_cli.insert_ucast_macs_remote(
                    ls_uuid, mac, port_ips, locator_uuid)
            else:
                ovsdb_cli.insert_ucast_macs_remote_and_locator(
                    ls_uuid, mac, port_ips, tunnel_ip)

    def reset_physical_port(self, lli, port, ip_mac_pairs, save=True):
        """Remove setting of raw of Physical_Port table in FOS switch OVSDB.

        ROWs with the same MAC address in Ucast_Macs_Local table and
        Ucast_Macs_Remote table are the target to extinct.

        :param lli: The local_link_information of the port.
        :type lli: list
        :param port: Context of the port.
        :type port: dictionary
        :param ip_mac_pairs: List of MAC(key) - IP(value) pairs of all FOS
                             switches.
        :type ip_mac_pairs: dictionary

        :returns: None
        """

        if lli:
            target = ip_mac_pairs[lli[0]['switch_id']]
            port_id = lli[0]['port_id']
            ovsdb_cli = ovsdb_writer.OVSDBWriter(target, self.ovsdb_port)
            ovsdb_cli.reset_physical_port(port_id)

        mac = port["mac_address"]

        for fossw_ip in self.fossw_ips:
            # All Ucast_Macs_Remote and Ucast_Macs_Local tables in all
            # FOS switches will delete the ROW with the MAC address.
            ovsdb_cli = ovsdb_writer.OVSDBWriter(fossw_ip, self.ovsdb_port)
            ovsdb_cli.delete_ucast_macs_local(mac)
            ovsdb_cli.delete_ucast_macs_remote(mac)
        if save:
            self.save_all_fossw()

    @utils.synchronized(_LOCK_NAME, external=True)
    def update_physical_port_with_lag(self, vni, llis, port, ip_mac_pairs,
                                      req_id):
        """Call update_physical_port for all physical swtich ports.

        :param vni: The segment ID of Neutron network which the port belongs
                     to.
        :type vni: integer
        :param llis: `local_link_information' of 'binding:profile' for the port
                     multiple physical ports(dict) should be given.
        :type llis: list
        :param port: Dictionary of the port.
        :type port: dictionary
        :param ip_mac_pairs: List of MAC(key) - IP(value) pairs of all FOS
                             switches.
        :type ip_mac_pairs: dictionary
        :param req_id: A request ID of network context which use for generate
                       new context.
        :type req_id: string

        :returns: None
        """

        for lli in llis:
            self.update_physical_port(
                vni, [lli], port, ip_mac_pairs, req_id)

    @utils.synchronized(_LOCK_NAME, external=True)
    def reset_physical_port_with_lag(self, llis, port, ip_mac_pairs):
        """Call reset_physical_port for all physical switch ports.

        :param llis: `local_link_information' of 'binding:profile' for the port
                     multiple physical ports(dict) should be given.
        :type llis: list
        :param port: Dictionary of the port.
        :type port: dictionary
        :param ip_mac_pairs: List of MAC(key) - IP(value) pairs of all FOS
                             switches.
        :type ip_mac_pairs: dictionary

        :returns: None
        """

        for lli in llis:
            self.reset_physical_port([lli], port, ip_mac_pairs)



# Copyright 2015-2017 FUJITSU LIMITED
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Implementation of Fujitsu ML2 Mechanism driver for ML2 Plugin."""

from networking_fujitsu._i18n import _
from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LI
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.common import utils

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants

from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import driver_api

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import importutils


LOG = logging.getLogger(__name__)
FUJITSU_DRIVER = 'networking_fujitsu.ml2.'
CFAB_DRIVER = FUJITSU_DRIVER + 'cfab.cfabdriver.CFABdriver'

VFAB_ID_DEFAULT = "default"
VFAB_ID_MIN = 1
VFAB_ID_MAX = 3000
_SUPPORTED_NET_TYPES = ['vlan']

ML2_FUJITSU_GROUP = "fujitsu_cfab"
ML2_FUJITSU = [
    cfg.StrOpt(
        'address', default='',
        help=_("The address of the C-Fabric to telnet to.")),
    cfg.StrOpt(
        'username', default='admin',
        help=_("The C-Fabric username to use.")),
    cfg.StrOpt(
        'password', default='admin', secret=True,
        help=_("The C-Fabric password to use.")),
    cfg.ListOpt(
        'physical_networks', default='',
        help=_("List of <physical_network>:<vfab_id> tuples specifying "
               "physical_network names and corresponding vfab ids.")),
    cfg.BoolOpt(
        'share_pprofile', default=False,
        help=_("Whether to share a C-Fabric pprofile among Neutron "
               "ports using the same VLAN ID.")),
    cfg.StrOpt(
        'pprofile_prefix', default='',
        help=_("The prefix string for pprofile name.")),
    cfg.BoolOpt(
        'save_config', default=True,
        help=_("Whether to save configuration."))]

cfg.CONF.register_opts(ML2_FUJITSU, ML2_FUJITSU_GROUP)


class CFABMechanismDriver(driver_api.MechanismDriver):
    """ML2 Mechanism driver for Fujitsu C-Fabric switches.

    This is the upper layer driver class that interfaces to lower layer (CLI)
    below.
    """

    def __init__(self):
        self._driver = None
        self._physical_networks = {}
        self._switch = None
        self.initialize()

    def initialize(self):
        """Initialize of variables needed by this class."""

        self._parse_physical_networks()
        self._switch = {
            'address': cfg.CONF.fujitsu_cfab.address,
            'username': cfg.CONF.fujitsu_cfab.username,
            'password': cfg.CONF.fujitsu_cfab.password
        }

        if not self._switch['address']:
            raise cfg.RequiredOptError(
                'address', cfg.OptGroup(ML2_FUJITSU_GROUP))
        self._driver = importutils.import_object(CFAB_DRIVER, cfg.CONF)

    def _parse_physical_networks(self):
        """Interpret physical_networks as physical_network:vfab_id entries."""

        method = "_parse_physical_networks"
        for entry in cfg.CONF.fujitsu_cfab.physical_networks:
            try:
                physical_network, vfab_id = entry.split(':')
            except ValueError:
                LOG.exception(_LE("Illegal physical_networks entry"))
                raise ml2_exc.MechanismDriverError(method=method)
            if not (vfab_id == VFAB_ID_DEFAULT or
                    VFAB_ID_MIN <= int(vfab_id) <= VFAB_ID_MAX):
                LOG.exception(_LE("Illegal VFAB in physical_networks entry"))
                raise ml2_exc.MechanismDriverError(method=method)
            self._physical_networks[physical_network] = vfab_id

    def _get_vfab_id(self, physical_network):
        """Get vfab_id corresponding to the physical_network."""

        try:
            vfab_id = self._physical_networks[physical_network]
        except KeyError:
            LOG.exception(
                _LE("Network not found in the configured physical network"))
            raise ml2_exc.MechanismDriverError(method="_get_vfab_id")
        return vfab_id

    @log_helpers.log_method_call
    def create_port_postcommit(self, mech_context):
        """Calls setup process for C-Fabric.

        Case1: Baremetal deploy
                   Setup VLAN for specified physical port.
        Case2: Otherwise
                   Associate the assigned MAC address to the portprofile.
        """
        port = mech_context.current
        network = mech_context.network
        if utils.is_baremetal(port):
            return

        if not is_supported(network):
            return

        method = 'create_port_postcommit'
        port_id = port['id']
        network_id = port['network_id']
        tenant_id = port['tenant_id']
        mac = port['mac_address']
        vfab_id = self._get_vfab_id(utils.get_physical_network(network))
        vlanid = utils.get_segmentation_id(network)

        try:
            self._driver.associate_mac_to_network(
                self._switch['address'],
                self._switch['username'],
                self._switch['password'],
                vfab_id,
                vlanid,
                mac
            )
        except Exception:
            LOG.exception(_LE("Failed to associate mac %s"), mac)
            raise ml2_exc.MechanismDriverError(method=method)

        LOG.info(
            _LI("created port (postcommit): port_id=%(port_id)s "
                "network_id=%(network_id)s tenant_id=%(tenant_id)s"),
            {'port_id': port_id,
             'network_id': network_id, 'tenant_id': tenant_id})

    @log_helpers.log_method_call
    def delete_port_postcommit(self, mech_context):
        """Calls cleanup process for C-Fabric.

        Case1: Baremetal deploy
                   Clear VLAN/LAG for specified physical port.
        Case2: Otherwise
                   Dissociate MAC address from the portprofile.
        """

        method = 'delete_port_postcommit'
        port = mech_context.current
        network = mech_context.network
        port_id = port['id']
        network_id = port['network_id']
        tenant_id = port['tenant_id']

        if utils.is_baremetal(port):
            if validate_baremetal_deploy(mech_context):
                params = self.get_physical_net_params(mech_context)
                try:
                    self.clear_vlan(params)
                except Exception:
                    LOG.exception(
                        _LE("Failed to clear VLAN(%s)."), params['vlanid'])
                    raise ml2_exc.MechanismDriverError(method=method)
        elif not is_supported(network):
            pass
        else:
            physical_network = utils.get_physical_network(network)
            vlanid = utils.get_segmentation_id(network)
            vfab_id = self._get_vfab_id(physical_network)
            mac = port['mac_address']

            try:
                self._driver.dissociate_mac_from_network(
                    self._switch['address'],
                    self._switch['username'],
                    self._switch['password'],
                    vfab_id,
                    vlanid,
                    mac)
            except Exception:
                LOG.exception(_LE("Failed to dissociate MAC %s"), mac)
                raise ml2_exc.MechanismDriverError(method=method)
        LOG.info(
            _LI("delete port (postcommit): port_id=%(p_id)s "
                "network_id=%(net_id)s tenant_id=%(tenant_id)s"),
            {'p_id': port_id, 'net_id': network_id, 'tenant_id': tenant_id})

    @log_helpers.log_method_call
    def setup_vlan(self, params):
        """Provision VLAN with specified port(s).

        This method will select driver's method.
        Case1: param['lag'] is True
            This method calls 'setup_vlan_with_lag' and clears VLAN and LAG.
        Case2: param['lag'] is False
            This method calls 'setup_vlan' and setup only VLAN.

        @param  params  a dictionary of the return value for
                        get_physical_net_params
        @return  None
        """

        target = 'setup_vlan_with_lag' if params['lag'] else 'setup_vlan'
        try:
            setup_method = getattr(self._driver, target)
            # This plugin supposes 1 C-Fabric(fabric_id) management.
            # Therefore, not to identify target IP address by using
            # switch_info(mac_address).
            LOG.info(_LI("call %(target)s.  params: %(params)s"),
                     {'target': target, 'params': params})
            setup_method(
                params['address'],
                params['username'],
                params['password'],
                params['vfab_id'],
                params['vlanid'],
                params['ports'],
                params['mac'],
            )
        except Exception:
            LOG.exception(_LE("Failed to setup VLAN(%s)"), params['vlanid'])
            raise ml2_exc.MechanismDriverError(method='setup_vlan')

    @log_helpers.log_method_call
    def clear_vlan(self, params):
        """Clear VLAN with specified port(s).

        This method will select driver's method.
        Case1: param['lag'] is True
            This method calls 'clear_vlan_with_lag' and clears VLAN and LAG.
        Case2: param['lag'] is False
            This method calls 'clear_vlan' and clears only VLAN.

        @param  params A dictionary of the return value for
                       get_physical_net_params
        @return  None
        """

        target = 'clear_vlan_with_lag' if params['lag'] else 'clear_vlan'
        try:
            clear_method = getattr(self._driver, target)
            # This plugin supposes 1 C-Fabric(fabric_id) management.
            # Therefore, not to identify target IP address by using
            # switch_info(mac_address).
            LOG.info(_LI("Call %(target)s.  params: %(params)s"),
                     {'target': target, 'params': params})
            clear_method(
                params['address'],
                params['username'],
                params['password'],
                params['vfab_id'],
                params['vlanid'],
                params['ports'],
                params['mac'],
            )
        except Exception:
            LOG.exception(_LE("Failed to clear VLAN(%s)"), params['vlanid'])
            raise ml2_exc.MechanismDriverError(target)

    @log_helpers.log_method_call
    def get_physical_net_params(self, mech_context):
        """Validate physical network parameters for baremetal deployment.

        Validates network & port params and returns dictionary.
        'local_link_information' is a dictionary from Ironic-port.  This value
        includes as follows:
            'switch_id': A string of switch's MAC address
                         This value is equal to 'chassis_id' from LLDP TLV.
            'port_id': A string of switch interface name.
                         This value is equal to 'port_id' from LLDP TLV.
            'switch_info': A string of switch name.
                         This value is equal to 'system_name' from LLDP TLV.

        @param  mech_context  a Context instance
        @return  A dictionary parameters for baremetal deploy
        """

        port = mech_context.current
        network = mech_context.network
        local_link_info = utils.get_physical_connectivity(port)
        return {
            "address": self._switch['address'],
            "username": self._switch['username'],
            "password": self._switch['password'],
            "ports": ','.join(p['port_id'] for p in local_link_info),
            "vfab_id": self._get_vfab_id(utils.get_physical_network(network)),
            "vlanid": utils.get_segmentation_id(network),
            "mac": port['mac_address'],
            "lag": utils.is_lag(local_link_info)
        }

    @log_helpers.log_method_call
    def bind_port(self, context):

        port = context.current
        vnic_type = port['binding:vnic_type']
        LOG.debug("Attempting to bind port %(port)s with vnic_type "
                  "%(vnic_type)s on network %(network)s",
                  {'port': port['id'], 'vnic_type': vnic_type,
                   'network': context.network.current['id']})

        if validate_baremetal_deploy(context):
            params = self.get_physical_net_params(context)
            segments = context.segments_to_bind
            self.setup_vlan(params)
            context.set_binding(segments[0][driver_api.ID],
                                portbindings.VIF_TYPE_OTHER, {},
                                status=constants.PORT_STATUS_ACTIVE)


def is_supported(network):
    """Validate network parameter(network_type and segmentation_id).

    @param a network object
    @return True if network_type is 'VLAN' and segmentation_id is included
            otherwise False
    """

    net_type = utils.get_network_type(network)
    if net_type not in _SUPPORTED_NET_TYPES:
        LOG.warning(_LW("Network type(%s) is not supported. Skip it."),
                    net_type)
        return False
    return True if utils.get_segmentation_id(network) else False


def validate_baremetal_deploy(mech_context):
    """Validate baremetal deploy.

    @param mech_context a context object
    @return True if enable to baremetal deploy otherwise False
    """

    port = mech_context.current
    network = mech_context.network
    if (utils.is_baremetal(port) and
       is_supported(network) and utils.get_physical_connectivity(port)):
        return True
    return False



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import mock
import socket
import time

from oslo_log import log as logging

from networking_fujitsu.ml2.common.ovsdb import base_connection

from neutron.tests import base

LOG = logging.getLogger(__name__)


class FakeDecodeClass(object):
    def __init__(self, fake_data):
        self.fake_data = fake_data

    def decode(self, utf8):
        return self.fake_data


class SocketClass(object):
    def __init__(self,
                 connect_error=None,
                 send_error=None,
                 recv_error=None,
                 rcv_data=None):
        self.connect_error = connect_error
        self.rcv_data = rcv_data
        self.send_error = send_error
        self.recv_error = recv_error

    def connect(self, ip_port):
        if self.connect_error:
            raise self.connect_error

    def send(self, data):
        if self.send_error:
            raise self.send_error
        return len(data)

    def recv(self, length):
        if self.recv_error:
            raise self.recv_error
        return self.rcv_data

    def close(self):
        pass


class TestBaseConnection(base.BaseTestCase):
    def setUp(self):
        super(TestBaseConnection, self).setUp()

        self.fake_ovsdb_ip = "192.168.1.1"
        self.fake_ovsdb_port = 6640

        self.sock = mock.patch('socket.socket').start()
        self.fake_ovsdb = base_connection.BaseConnection(
            self.fake_ovsdb_ip, self.fake_ovsdb_port)
        self.fake_ovsdb.s = mock.patch('socket.socket').start()
        self.op_id = 'abcd'
        self.fake_message = {'id': self.op_id,
                             'fake_key': 'fake_value'}

        self.fake_ovsdb.responses = [self.fake_message]

        self.fake_ip = 'fake_ip'

    def test_init(self):
        fakesocket = SocketClass()
        with mock.patch.object(base_connection.LOG, 'debug') as logger_call, \
                mock.patch.object(socket, 'socket',
                                  return_value=fakesocket):
            self.fake_ovsdb.__init__(self.fake_ovsdb_ip, self.fake_ovsdb_port)
            self.assertTrue(logger_call.called)
            self.assertTrue(self.sock.called)

    def test_init_with_socket_error(self):
        fakesocket = SocketClass(socket.error)
        with mock.patch.object(base_connection.LOG, 'exception') as logger_exc, \
                mock.patch.object(base_connection.LOG, 'warning') as logger_warn, \
                mock.patch.object(socket, 'socket', return_value=fakesocket) as sock_connect, \
                mock.patch.object(time, 'sleep'):
            self.assertRaises(socket.error, base_connection.BaseConnection,
                              self.fake_ovsdb_ip, self.fake_ovsdb_port)
            self.assertTrue(logger_warn.called)
            self.assertTrue(logger_exc.called)
            self.assertTrue(sock_connect.called)

    def test_init_with_timeout(self):
        fakesocket = SocketClass(socket.timeout)
        with mock.patch.object(base_connection.LOG, 'exception') as logger_exc, \
                mock.patch.object(base_connection.LOG, 'warning') as logger_warn, \
                mock.patch.object(socket, 'socket', return_value=fakesocket) as sock_connect, \
                mock.patch.object(time, 'sleep'):
            self.assertRaises(socket.timeout, base_connection.BaseConnection,
                              self.fake_ovsdb_ip, self.fake_ovsdb_port)
            self.assertTrue(logger_warn.called)
            self.assertTrue(logger_exc.called)
            self.assertTrue(sock_connect.called)

    def test_send(self):
        with mock.patch.object(self.fake_ovsdb.socket, 'send',
                               return_value=1):
            return_value = self.fake_ovsdb.send(mock.Mock())
            self.assertEqual(return_value, True)

    def test_send_exception(self):
        with mock.patch.object(self.fake_ovsdb.socket, 'send',
                               side_effect=Exception) as send:
            with mock.patch.object(base_connection.LOG,
                                   'exception'):
                with mock.patch.object(self.fake_ovsdb,
                                       'disconnect') as mock_disconnect:
                    with mock.patch.object(base_connection.LOG,
                                           'warning'):
                        self.fake_ovsdb.send(mock.Mock())
                        self.assertTrue(send.called)
                        self.assertTrue(mock_disconnect.called)

    def test_disconnect(self):
        with mock.patch.object(self.fake_ovsdb.socket, 'close') as sock_close:
            self.fake_ovsdb.disconnect()
            self.assertTrue(sock_close.called)



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from networking_fujitsu.ml2.common import type_vxlan

from neutron.plugins.common import constants as p_const
from neutron.tests.unit.plugins.ml2.drivers import base_type_tunnel
from neutron.tests.unit import testlib_api


class TestTypeVxlan(base_type_tunnel.TunnelTypeTestMixin,
                    testlib_api.SqlTestCase):
    DRIVER_MODULE = type_vxlan
    DRIVER_CLASS = type_vxlan.TypeVxlan
    TYPE = p_const.TYPE_VXLAN

    def add_endpoint(self, ip=base_type_tunnel.TUNNEL_IP_ONE,
                     host=base_type_tunnel.HOST_ONE,
                     port=4789):
        return self.driver.add_endpoint(ip, host, port)

    def test_add_endpoint(self):
        endpoint = super(TestTypeVxlan, self).test_add_endpoint()
        self.assertEqual(4789, endpoint.udp_port)

    def test_get_endpoints(self):
        self.add_endpoint()
        self.add_endpoint(base_type_tunnel.TUNNEL_IP_TWO,
                          base_type_tunnel.HOST_TWO)

        endpoints = self.driver.get_endpoints()
        for endpoint in endpoints:
            if endpoint['ip_address'] == base_type_tunnel.TUNNEL_IP_ONE:
                self.assertEqual(4789, endpoint['udp_port'])
                self.assertEqual(base_type_tunnel.HOST_ONE, endpoint['host'])
            elif endpoint['ip_address'] == base_type_tunnel.TUNNEL_IP_TWO:
                self.assertEqual(4789, endpoint['udp_port'])
                self.assertEqual(base_type_tunnel.HOST_TWO, endpoint['host'])

    def test_db_get_endpoint_by_ip(self):
        self.add_endpoint()
        endpoint = self.driver.db_get_endpoint_by_ip(
            base_type_tunnel.TUNNEL_IP_ONE)
        self.assertEqual(4789, endpoint['udp_port'])
        self.assertEqual(base_type_tunnel.HOST_ONE, endpoint['host'])

    def test_db_delete_endpoint(self):
        self.add_endpoint()

        self.assertIsNone(self.driver.db_delete_endpoint(
            base_type_tunnel.TUNNEL_IP_ONE))
        # Get all the endpoints and verify its empty
        endpoints = self.driver.get_endpoints()
        self.assertNotIn(base_type_tunnel.TUNNEL_IP_ONE, endpoints)

    def test_db_get_endpoint_ip_by_host(self):
        self.add_endpoint()
        endpoint_ip = self.driver.db_get_endpoint_ip_by_host(
            base_type_tunnel.HOST_ONE)
        self.assertEqual(base_type_tunnel.TUNNEL_IP_ONE, endpoint_ip)
